/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * "Portions Copyright (c) 1999 Apple Computer, Inc.  All Rights
 * Reserved.  This file contains Original Code and/or Modifications of
 * Original Code as defined in and that are subject to the Apple Public
 * Source License Version 1.0 (the 'License').  You may not use this file
 * except in compliance with the License.  Please obtain a copy of the
 * License at http://www.apple.com/publicsource and read it before using
 * this file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License."
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#define OID_MOUNTPASSNUMBER			"mountPassNumber"
#define OID_MOUNTDUMPFREQUENCY			"mountDumpFrequency"
#define OID_MOUNTOPTION				"mountOption"
#define OID_MOUNTTYPE				"mountType"
#define OID_MOUNTDIRECTORY			"mountDirectory"
#define OID_AUTOMOUNT				"automount"
#define OID_LOGINSHELL				"loginShell"
#define OID_HOMEDIRECTORY			"homeDirectory"
#define OID_GECOS				"gecos"
#define OID_GIDNUMBER				"gidNumber"
#define OID_UIDNUMBER				"uidNumber"
#define OID_USERPASSWORD			"userPassword"
#define OID_MODIFYTIMESTAMP			"modifyTimestamp"
#define OID_OBJECTCLASS				"objectClass"
#define OID_NISNETGROUP				"nisNetgroup"
#define OID_IPHOST				"ipHost"
#define OID_DESCRIPTION				"description"
#define OID_ONCRPC				"oncRpc"
#define OID_IPPROTOCOL				"ipProtocol"
#define OID_IPSERVICE				"ipService"
#define OID_POSIXGROUP				"posixGroup"
#define OID_POSIXACCOUNT			"posixAccount"
#define OID_CN					"cn"
#define OID_IPNETWORK				"ipNetwork"
#define OID_MOUNT				"mount"
#define OID_RFC822MAILGROUP			"rfc822MailGroup"
#define OID_TTL					"ttl"
#define OID_BOOTABLEDEVICE			"bootableDevice"
#define OID_CRYPT				"{CRYPT}"
#define OID_MAIL				"mail"
#define OID_UID					"uid"
#define OID_IEEE802DEVICE			"ieee802Device"
#define OID_IPSERVICEPROTOCOL			"ipServiceProtocol"
#define OID_IPSERVICEPORT			"ipServicePort"
#define OID_NISNETGROUPTRIPLE			"nisNetgroupTriple"
#define OID_MEMBERNISNETGROUP			"memberNisNetgroup"
#define OID_MEMBERUID				"memberUid"
#define OID_IPHOSTNUMBER			"ipHostNumber"
#define OID_ONCRPCNUMBER			"oncRpcNumber"
#define OID_IPPROTOCOLNUMBER			"ipProtocolNumber"
#define OID_BOOTFILE				"bootFile"
#define OID_BOOTPARAMETER			"bootParameter"
#define OID_MACADDRESS				"macAddress"
#define OID_IPNETMASKNUMBER			"ipNetmaskNumber"
#define OID_IPNETWORKNUMBER			"ipNetworkNumber"
#define OID_LPRHOST				"lprHost"
#define OID_LPRTYPE				"lprType"
#define OID_LPRQUEUE				"lprQueue"
#define OID_LPRPRINTER				"lprPrinter"
